# Copyright 2022-present Lenovo
# Confidential and Proprietary
import json
import os
import shutil
import sys
import uuid
from collections import defaultdict

from datasource_process import (
    convert_and_generate_thumbnail,
    convert_and_generate_thumbnail_core,
    upload_pics
)
from utils import (
    parse_xml,
    reformat_and_save_xml,
    remove_file,
    get_dest_filename,
    delete_last_line
)


class DatasetTransformer(object):
    def __init__(self, working_dir):
        self.working_dir = working_dir
        self.file_info = {
            "size_b": 0,
            "label_num": 0,
            "total_num": 0,
            "label_statistics": defaultdict(int)
        }

    def _check_voc_file(self, import_path):
        import zipfile
        namelist = zipfile.ZipFile(import_path).namelist()
        dirs = {i.split(os.sep)[0] for i in namelist}
        check_dir = {'ImageSets', 'Annotations', 'labelmap.txt', 'JPEGImages'}
        return dirs == check_dir

    def _init_voc_path(self, origin_voc_ds):
        self.annotations_dir = os.path.join(
            self.working_dir, 'origin', 'Annotations')
        self.images_dir = os.path.join(
            self.working_dir, 'origin', 'JPEGImages')
        self.thumbnail_dir = os.path.join(self.working_dir, 'thumbnail')
        self.unlabel_dir = os.path.join(self.working_dir, 'unlabel')
        self.file_json_path = os.path.join(
            self.working_dir, "origin", "file_info.json")
        self.record_file_path = os.path.join(
            self.working_dir, 'record_file.txt')
        self.category_file = os.path.join(
            self.working_dir, 'origin', 'category.txt')
        self.temp_dest_dir = upload_pics(
            origin_voc_ds, str(uuid.uuid1()), self.working_dir)
        self.origin_images = os.path.join(self.temp_dest_dir, "JPEGImages")
        self.origin_xmls = os.path.join(self.temp_dest_dir, "Annotations")
        self.labelmap_path = os.path.join(
            self.temp_dest_dir, "labelmap.txt")

    def _update_category_file(self, labels):
        with open(self.category_file, "r+") as f:
            data = f.read().split('\n')[:-1]
            write_labels = [i for i in labels if i not in data]
            if not write_labels:
                # 减去__background__
                return len(data) - 1
            write_content = "\n".join(write_labels)
            f.write(f"{write_content}\n")
        return len(write_labels) + len(data) - 1

    def _write_labelmap(self):
        if not os.path.isfile(self.labelmap_path):
            return
        with open(self.labelmap_path) as f:
            raw = f.readlines()[2:]
            labels = {i[:i.find(":")] for i in raw}
        if not labels:
            return
        for label in labels:
            if label not in self.file_info["label_statistics"]:
                self.file_info["label_statistics"][label]
        written_labels = self._update_category_file(labels)
        self.file_info["label_num"] = written_labels
        self._write_json_file(self.file_json_path, self.file_info)

    def _delete_file_record(self, img_path):
        if not os.path.isfile(self.record_file_path) or not img_path:
            return
        with open(self.record_file_path, "r") as r:
            lines = r.readlines()
        with open(self.record_file_path, 'w') as w:
            for line in lines:
                if img_path not in line:
                    w.write(line)

    def _convert_img_data(self, xml_path):
        try:
            dst_img_path = ''
            # 获取图片的原始路径和拷贝后的路径
            filename, labeled = parse_xml(xml_path)
            src_img_path = os.path.join(
                self.temp_dest_dir, "JPEGImages", filename)
            img = os.path.split(src_img_path)[1]
            img_name, img_type = os.path.splitext(img)
            temp_dest_path = os.path.join(self.images_dir, img) if labeled \
                else os.path.join(self.unlabel_dir, img)
            ok, temp_dst_img_path = get_dest_filename(
                temp_dest_path, self.thumbnail_dir)
            if not ok:
                return False
            dst_img_path = f"{os.path.splitext(temp_dst_img_path)[0]}.jpg"
            dst_img_name = os.path.split(dst_img_path)[1]
            dst_basename = os.path.splitext(dst_img_name)[0]
            ok, thumbnail_image_path = convert_and_generate_thumbnail_core(
                src_img_path, dst_img_path, self.thumbnail_dir,
                self.record_file_path)
            if not ok:
                return False
            if labeled:
                # 转xml格式并拷贝到loaction/Annotations/目录下, 返回这张图片所有label
                labels = reformat_and_save_xml(
                    xml_path,
                    os.path.join(self.annotations_dir, f"{dst_basename}.xml"),
                    os.path.split(dst_img_path)[1])
                for label in labels:
                    self.file_info["label_statistics"][label] += 1
                # 把labels写入category.txt, 返回目前category.txt里的label数量
                label_num = self._update_category_file(set(labels))
                self.file_info["label_num"] = label_num
        except Exception as e:
            print(f"[UPLOAD_ERROR]: Import {img} error: {e}")
            self._delete_file_record(dst_img_path)
            remove_file([dst_img_path, thumbnail_image_path])
            return False
        else:
            self.file_info["size_b"] += os.path.getsize(dst_img_path)
            self.file_info["total_num"] += 1
            self._write_json_file(self.file_json_path, self.file_info)
            return True

    def import_voc_dataset(self, origin_voc_ds):
        if not self._check_voc_file(origin_voc_ds):
            print(f"[ERROR]: The VOC import file: {origin_voc_ds} is invalid")
            sys.exit(1)
        self._init_voc_path(origin_voc_ds)
        success_num = 0
        failed_num = 0
        for curDir, dirs, files in os.walk(self.origin_xmls):
            if not files:
                continue
            for xml in files:
                success = self._convert_img_data(os.path.join(curDir, xml))
                if success:
                    success_num += 1
                else:
                    failed_num += 1
        self._write_labelmap()
        if os.path.exists(self.temp_dest_dir):
            shutil.rmtree(self.temp_dest_dir)
        print("[IMPORT_FINISH]: The successful import statistic is %d. "
              "The failed import statistic is %d."
              % (success_num, failed_num))

    def import_coco_dataset(self, origin_coco_ds):  # noqa: C901
        temp_name = str(uuid.uuid1())
        annotations_dir = os.path.join(
            self.working_dir, 'origin', 'Annotations')  # 最后要存放标注信息的文件夹
        images_dir = os.path.join(
            self.working_dir, 'origin', 'JPEGImages')  # 最后要存放已标注图片的文件夹
        thumbnail_dir = os.path.join(self.working_dir, 'thumbnail')  # 缩略图的文件夹
        unlabel_dir = os.path.join(self.working_dir, 'unlabel')  # 未标注图片的文件夹
        record_file_path = os.path.join(
            self.working_dir, 'record_file.txt')  # 记录图片文件路径
        category_file = os.path.join(
            self.working_dir, 'origin', 'category.json')  # 存放label信息的文件
        dest_dir = upload_pics(
            origin_coco_ds, temp_name, self.working_dir)  # 临时存放解压的文件夹
        file_json_path = os.path.join(
            self.working_dir, "origin", "file_info.json")

        if set(os.listdir(dest_dir)) != {"annotations", "images"}:
            print("[ERROR]: The ZIP file is invalid!")
            sys.exit(1)
        origin_annotations = os.path.join(dest_dir, "annotations")
        origin_images = os.path.join(dest_dir, "images")
        origin_annotations_list = os.listdir(origin_annotations)
        field_dict = {}
        success_images_nums = 0
        failed_images_nums = 0
        # 根据注释信息导入图片数据，剩下的图片是未标注的图片
        for origin_annotations_file in origin_annotations_list:
            if not origin_annotations_file.endswith(".json"):
                print("[Warning]: A file that does not end with .json was "
                      "found %s" % os.path.basename(origin_annotations_file))
                continue
            with open(os.path.join(
                    origin_annotations, origin_annotations_file),
                    'r', encoding='utf-8') as f:
                all_annotation_data = json.load(f)
            # Write label information
            categories_list = all_annotation_data.get("categories", [])
            if categories_list:
                for category in categories_list:
                    self.file_info["label_statistics"][
                        category["name"]] = 0
                new_label_dict = self._wrtie_category(
                    category_file, categories_list)
                ok = self._write_json_file(file_json_path, self.file_info)
                if ok:
                    print("[SUCCESS]: Generate the file_info file.")
                else:
                    print("[ERROR]: Generate the file_info file failed, the "
                          "contents of the document are as "
                          "follows:%s" % self.file_info)
                    continue
            # Write the annotation information and move the image
            images_data_list = all_annotation_data.get("images", [])
            success_images_nums += len(images_data_list)
            new_images_dict = {}
            for images in images_data_list:  # type:dict
                image_id = images.pop("id")
                new_images_dict[image_id] = images
            annotations_list = all_annotation_data.get("annotations", [])
            thumbnail_image_path = None
            for annotations in annotations_list:  # type:dict
                image_id = annotations.pop("image_id")
                image_name = new_images_dict[image_id]["file_name"]
                if image_name in field_dict:
                    continue
                src_image_file = os.path.join(origin_images, image_name)
                image_name_jpg = "%s.jpg" % os.path.splitext(
                    os.path.basename(image_name))[0]
                dst_image_file = os.path.join(images_dir, image_name_jpg)
                # 1. save image, thumbnail and write record_file.txt
                if os.path.exists(src_image_file):
                    ok, dst_image_file = get_dest_filename(
                        dst_image_file, thumbnail_dir)
                    if not ok:
                        field_dict[image_name] = True
                        failed_images_nums += 1
                        success_images_nums -= 1
                        continue
                    image_name = os.path.basename(dst_image_file)
                    new_images_dict[image_id]["file_name"] = image_name
                    ok, thumbnail_image_path = \
                        convert_and_generate_thumbnail_core(
                            src_image_file, dst_image_file,
                            thumbnail_dir, record_file_path)
                    os.remove(src_image_file)
                    if not ok:
                        field_dict[image_name] = True
                        failed_images_nums += 1
                        success_images_nums -= 1
                        continue
                # 2. save annotations
                ann_temp_filename = "%s.json" % os.path.splitext(
                    os.path.basename(dst_image_file))[0]
                annotations_file = os.path.join(
                    annotations_dir, ann_temp_filename)
                del annotations["id"]
                del annotations["attributes"]
                label_name = new_label_dict[annotations["category_id"]]
                annotations["category_id"] = new_label_dict[label_name]
                if os.path.exists(annotations_file):
                    with open(annotations_file, 'r', encoding='utf-8') as rf:
                        annotation_data = json.load(rf)
                    annotation_data["annotations"].append(annotations)
                else:
                    annotation_data = {"images": [new_images_dict[image_id]],
                                       "annotations": [annotations]}
                self.file_info["label_statistics"][label_name] += 1
                ok = self._write_json_file(annotations_file, annotation_data)
                if ok:
                    print("[IMPORT_ANNOTATIONS]: Import annotations "
                          "%s success!" % image_name)
                else:
                    remove_file([dst_image_file, thumbnail_image_path])
                    delete_last_line(record_file_path)
                    field_dict[image_name] = True
                    print("[IMPORT_ANNOTATIONS]: Import annotations "
                          "%s faield!" % image_name)
                    continue
                ok = self._write_json_file(file_json_path, self.file_info)
                if ok:
                    print("[SUCCESS]: Update the file_info file.")
                else:
                    remove_file([dst_image_file, thumbnail_image_path])
                    delete_last_line(record_file_path)
                    field_dict[image_name] = True
                    print("[ERROR]: Update the file_info file failed, the "
                          "contents of the document are as "
                          "follows:%s" % self.file_info)

        # import unlabel images
        print("Begin convert unlabel picture and generate thumbnail...")
        success_num, failed_num = convert_and_generate_thumbnail(
            origin_images, unlabel_dir, thumbnail_dir, record_file_path)
        print("End of unlabeled image conversion: success %s; "
              "faield %s" % (success_num, failed_num))
        failed_images_nums += failed_num
        # write file_info file: Only label_num can be used directly,
        # and other data needs to be combined with the database.

        if os.path.exists(dest_dir):
            shutil.rmtree(dest_dir)
        print("[IMPORT_FINISH]: The successful import statistic is %d. "
              "The failed import statistic is %d."
              % (success_images_nums, failed_images_nums))
        if field_dict:
            print("All failed pictures are below:%s" % [i for i in field_dict])

    def _wrtie_category(self, category_file, cvat_categories,
                        new_label_dict=None):

        """
        Args:
            category_file[string]: category file
            cvat_categories[list]: List of categories to add
            new_label_dict[dict]: Label and Label ID mapping
                                    {"circle":0,0:"circle"}
        Returns:
            new_label_dict
        """
        if not new_label_dict:
            new_label_dict = dict()
            for label_dict in cvat_categories:
                new_label_dict[label_dict["name"]] = label_dict["id"]
                new_label_dict[label_dict["id"]] = label_dict["name"]
        extend_label_list = []  # 要加进去的新的category
        with open(category_file, 'r', encoding='utf-8') as f:
            category_file_info = json.load(f)
        lico_categories_list = category_file_info["categories"]
        max_category_id = 0
        lico_categories_dict = {}
        if lico_categories_list:
            for category_dict in lico_categories_list:
                lico_categories_dict[
                    category_dict["name"]] = category_dict
                if max_category_id < category_dict["id"]:
                    max_category_id = category_dict["id"]
            for category_dict in cvat_categories:
                category_name = category_dict["name"]
                if category_name in lico_categories_dict:
                    new_label_dict[category_name] = \
                        lico_categories_dict[category_name]["id"]
                else:
                    max_category_id += 1
                    new_label_dict[category_name] = max_category_id
                    category_dict["id"] = max_category_id
                    extend_label_list.append(category_dict)
        else:
            extend_label_list = cvat_categories
        category_file_info["categories"].extend(extend_label_list)
        self.file_info["label_num"] = len(category_file_info["categories"])
        ok = self._write_json_file(category_file, category_file_info)
        if ok:
            print("[LABEL]: Import label success!")
        else:
            print("[LABEL]: Import label faield! Program exits!")
            sys.exit(1)

        return new_label_dict

    def _write_json_file(self, file, data):
        try:
            with open(file, 'w', encoding='utf-8') as f:
                json.dump(data, f)
            return True
        except Exception:
            print("Save parameter to json file failed! "
                  "Json file: %s" % os.path.basename(file))
            return False
