# Copyright 2022-present Lenovo
# Confidential and Proprietary


"""
Example:
python datasource_process.py
    --location='/home/hpcadmin/ai_stdio/datasource/test'
    --label='dog' --image_path='/home/hpcadmin/test.zip'
"""

import argparse
import json
import os
import re
import shutil
import tarfile
import uuid
import zipfile
import imghdr
from collections import defaultdict

from PIL import Image

from utils import get_dest_filename

ACCEPTABLE_FILE_PATTERNS = {
    '.jpg', '.png', '.jpeg', '.bmp', '.gif'}


def convert_and_generate_thumbnail(
        image_path, origin_dir, thumb_dir, record_file_path):
    success_num = 0
    failed_num = 0
    if not os.path.isdir(image_path):
        print('[UPLOAD_ERROR]: The image path is incorrect')
        return success_num, failed_num
    try:
        for root, dirs, files in os.walk(image_path, topdown=False):
            failed_num += len(files)
            for file_name in files:
                srcfile = os.path.join(root, file_name)
                picture, ext = os.path.splitext(file_name)
                if ext.lower() in ACCEPTABLE_FILE_PATTERNS:
                    dstfile = os.path.join(origin_dir, picture + '.jpg')
                    ok, dstfile = get_dest_filename(dstfile, thumb_dir)
                    if not ok:
                        continue
                    flag, _ = convert_and_generate_thumbnail_core(
                        srcfile, dstfile, thumb_dir, record_file_path)
                    if flag:
                        success_num += 1
                        failed_num -= 1
    except Exception as e:
        print("[UPLOAD_ERROR]: Convert picture error, reason: %s" % str(e))
    return success_num, failed_num


def convert_and_generate_thumbnail_core(
        srcfile, dstfile, thumb_dir, record_file_path):
    thumbnail_image_path = ''
    img_name, ext = os.path.splitext(os.path.basename(srcfile))
    img_type = f'.{imghdr.what(srcfile)}'
    if img_type not in ACCEPTABLE_FILE_PATTERNS:
        print(f'[UPLOAD_ERROR]: Import {img_name}{ext} error:',
              f'the real format of {img_name} is {img_type}')
        return False, thumbnail_image_path
    flag = copy_or_convert_image(srcfile, ext.lower(), dstfile)
    if flag:
        flag, thumbnail_image_path = generate_thumbnail(
            thumb_dir, dstfile, srcfile, record_file_path)
    return flag, thumbnail_image_path


def copy_or_convert_image(srcfile, ext, dstfile):
    pic_path = os.path.split(srcfile)[1]
    try:
        if ext == '.jpg':
            shutil.copyfile(srcfile, dstfile)
        else:
            with Image.open(srcfile) as image:
                if image.mode in ['P', 'RGBA']:
                    image.convert('RGB').save(dstfile)
                else:
                    image.save(dstfile)
        print('[CONVERT]: Convert %s successfully.' % pic_path)
        return True
    except Exception:
        if os.path.exists(dstfile):
            os.remove(dstfile)
        print('[UPLOAD_ERROR]: The image format cannot be converted .JPG : %s.'
              % pic_path)
    return False


def generate_thumbnail(thumb_dir, dstfile, srcfile, record_file_path):
    """Generate thumbnails"""
    pic_path = os.path.split(srcfile)[1]
    thumbnail_size = [104, 74]
    filepath, tempfilename = os.path.split(dstfile)
    picturename, ex = os.path.splitext(tempfilename)
    save_path = os.path.join(thumb_dir, picturename + ".gif")
    try:
        with Image.open(dstfile) as image:
            # Generate thumbnail
            image.thumbnail(thumbnail_size, Image.ANTIALIAS)
            # Add black background
            b = Image.new(mode="RGBA", size=thumbnail_size)
            b.paste(image, (int((b.size[0] - image.size[0]) / 2),
                            int((b.size[1] - image.size[1]) / 2)))
        b.save(save_path, "GIF")
        print('[THUMBNAIL]:  %s thumbnail created successfully!' % pic_path)
        with open(record_file_path, "a+") as f:
            f.write(dstfile + '\n')
            f.flush()
        return True, save_path
    except Exception:
        if os.path.exists(dstfile):
            os.remove(dstfile)
        if os.path.exists(save_path):
            os.remove(save_path)
        print('[UPLOAD_ERROR]: Thumbnail creation failed : %s ' % pic_path)
    return False, save_path


def parse_args():
    """Parse input arguments."""
    parser = argparse.ArgumentParser(description='Datasource process script')
    parser.add_argument('--location', dest='location',
                        help='Datasource project path',
                        required=True, type=str)
    parser.add_argument('--image_path', dest='image_path',
                        help='Upload pictures path', type=str)
    parser.add_argument('--label', dest='label',
                        help='Image classification label',
                        default='')

    # for text dataset
    parser.add_argument('--file_path', dest='file_path',
                        help='Text classification import file path',
                        default='')
    parser.add_argument('--max_lines', dest='max_lines',
                        help='Text classification file max lines limit',
                        default=50000)

    # for import dataset
    parser.add_argument('--import_path', dest='import_path',
                        help='Import file path', type=str)
    parser.add_argument('--import_type', dest='import_type',
                        help='Import dataset type', type=str)

    return parser.parse_args()


def get_temp_dir(location, temp_name):
    dest_dir = os.path.join(location, 'temp', temp_name)
    temp_dir = os.path.join(location, 'temp')
    if not os.path.exists(temp_dir):
        os.mkdir(temp_dir)
    if not os.path.exists(dest_dir):
        os.mkdir(dest_dir)
    return dest_dir


def upload_pics(image_path, temp_name, location):
    try:
        dest_dir = get_temp_dir(location, temp_name)
        if os.path.isdir(image_path):
            shutil.rmtree(dest_dir)
            shutil.copytree(image_path, dest_dir)
        elif zipfile.is_zipfile(image_path):
            print("Decompress zip file...")
            with zipfile.ZipFile(image_path) as zip:
                for name in zip.namelist():
                    zip.extract(name, path=dest_dir)
        elif tarfile.is_tarfile(image_path):
            print("Decompress tar file...")
            with tarfile.open(image_path) as tar:
                tar.extractall(path=dest_dir)
        elif os.path.isfile(image_path):
            shutil.copy(image_path, dest_dir)
        else:
            print("[UPLOAD_ERROR]: Unsupported image "
                  "path %s, please check!" % (image_path))
            exit(1)
    except Exception as e:
        print("[UPLOAD_ERROR]: Upload picture error" + str(e))
    return dest_dir


def get_parameter_values(filename_path, counts):
    try:
        with open(filename_path, 'r') as file_read:
            total_num = 0
            for line in file_read:
                re_result = line.split('\t', 1)
                labels = re_result[0]
                counts[labels] += 1
                total_num += 1
    except Exception:
        print(f"The {filename_path} format is wrong")
    return total_num


def get_dir_size_b(filename_path):
    if os.path.isfile(filename_path):
        return os.path.getsize(filename_path)
    else:
        return 0


def copy_content_to_default_file(
        import_file_path, location, max_lines):
    filename_path = os.path.join(location, "origin/train.tsv")
    file_num = get_default_file_lines(location)
    success_num = 0
    failed_num = 0
    with open(import_file_path, 'r') as file_read:
        with open(filename_path, 'ab') as file_add_cont:
            for line in file_read:
                try:
                    if file_num + success_num >= int(max_lines):
                        print('The number of lines in the file exceeds '
                              'the maximum limit')
                        break
                    category, sample = line.strip().split('\t', 1)
                    if re.search(r"\s", category):
                        failed_num += 1
                        print("line_num %d category format is invalid."
                              % (success_num + failed_num))
                        continue
                    file_add_cont.write(line.encode("utf-8"))
                    success_num += 1
                except Exception:
                    failed_num += 1
                    print("line_num %d content format is invalid."
                          % (success_num + failed_num))
    print(f"[UPLOAD_FINISH]: The successful upload statistic "
          f"is {success_num}, The failed upload statistic is {failed_num}.")


def get_default_file_lines(location):
    file_json_path = os.path.join(location, "origin/file_info.json")
    if not os.path.exists(file_json_path):
        print(f"{file_json_path} file path Invalid")
        exit(1)
    try:
        with open(file_json_path, 'r') as file_read:
            file_json_dict = json.load(file_read)
            file_num = file_json_dict.get('total_num')
            if file_num >= int(max_lines):
                print(f"The {file_json_path} total lines exceeds the limit")
                exit(1)
    except Exception:
        print(f"{file_json_path} read error")
    return file_num


if __name__ == '__main__':  # noqa: C901
    args = parse_args()
    location = args.location
    if args.import_path:
        from dataset_transformer import DatasetTransformer

        dataset_trans = DatasetTransformer(location)
        if args.import_type == 'voc':
            dataset_trans.import_voc_dataset(args.import_path)
        elif args.import_type == 'coco':
            dataset_trans.import_coco_dataset(args.import_path)
        else:
            print("Unsupported import dataset type.")
            exit(1)
    elif args.file_path:
        counts = defaultdict(int)
        import_file_path = args.file_path
        max_lines = args.max_lines
        filename_path = os.path.join(location, "origin/train.tsv")
        copy_content_to_default_file(import_file_path, location, max_lines)
        total_num = get_parameter_values(filename_path, counts)
        file_size = get_dir_size_b(filename_path)
        file_json_dict = {
            "size_b": file_size,
            "label_num": len(counts),
            "total_num": total_num,
            "label_statistics": counts
        }
        try:
            file_json_path = os.path.join(location, "origin/file_info.json")
            with open(file_json_path, 'w') as file_write:
                json.dump(file_json_dict, file_write)
        except Exception:
            print("Save parameter to json file failed")
    else:
        image_path = args.image_path
        label = args.label
        temp_name = str(uuid.uuid1())

        print("Run datasource process script start...")
        if not all([location, image_path, temp_name]):
            print("[UPLOAD_ERROR]: Args error, please check!")
            exit(1)

        if label:
            origin_dir = os.path.join(location, "origin", label)
            thumbnail_dir = os.path.join(location, "thumbnail", label)
            record_file_path = os.path.join(
                location, "thumbnail", label + "_record_file.txt")
        else:
            origin_dir = os.path.join(location, "unlabel")
            thumbnail_dir = os.path.join(location, "thumbnail")
            record_file_path = os.path.join(location, "record_file.txt")
        try:
            dest_dir = upload_pics(image_path, temp_name, location)

            print("Begin convert picture and generate thumbnail...")
            success_num, failed_num = \
                convert_and_generate_thumbnail(
                    dest_dir, origin_dir, thumbnail_dir,
                    record_file_path)
            print("[UPLOAD_FINISH]: The successful upload statistic is %d. "
                  "The failed upload statistic is %d."
                  % (success_num, failed_num))
        except Exception:
            print("[UPLOAD_ERROR]: Upload picture error")
        finally:
            dest_dir = os.path.join(location, 'temp', temp_name)
            if os.path.exists(dest_dir):
                shutil.rmtree(dest_dir)
