# Copyright 2021-present Lenovo
# Confidential and Proprietary

import csv
import json
from itertools import chain

import cv2
import numpy as np
import pycuda.autoinit  # noqa: F401
import pycuda.driver as cuda
import tensorrt as trt


def per_image_standardization(img):
    mean = np.mean(img)
    std = np.std(img)
    count = 1
    for size in img.shape:
        count = count * size
    adj_std = max(std, 1.0 / np.sqrt(count))
    img = (img - mean) / adj_std
    return img


def load_image(image_file, image_size):
    np_img = cv2.imread(image_file, flags=1)
    np_img = np_img.astype(np.float32)
    np_img = cv2.resize(np_img, (image_size, image_size))
    np_img = per_image_standardization(np_img)
    np_img = np_img[:, :, [2, 1, 0]]  # BGR2RGB
    return np_img.ravel()


class ArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.__str__()
        return json.JSONEncoder.default(self, obj)


def convert_to_list(dict_val):
    list_key = list(dict_val)
    list_val = list(dict_val.values())
    list_result = list(chain.from_iterable([zip(list_key, list_val)]))
    return list_result


def output_to_csv(filename, first_row, other_rows):
    with open(filename, 'w') as f:
        csv_writer = csv.writer(f)
        csv_writer.writerow(first_row)
        for row in other_rows:
            csv_writer.writerow(row)


def output_to_file(contents, filename):
    if isinstance(contents, dict):
        contents = json.dumps(contents, cls=ArrayEncoder)
        with open(filename, 'w') as f:
            f.write(contents)
    elif isinstance(contents, list):
        with open(filename, 'w') as f:
            for v in contents:
                f.write(v)
    elif isinstance(contents, str):
        with open(filename, 'w') as f:
            f.write(contents)


def load_labels(label_file):
    labels_dict = {}
    with open(label_file) as f:
        for line in f:
            key, value = line.rstrip('\n').split(':')
            labels_dict[key] = value
    return labels_dict


def do_inference(context, bindings, inputs, outputs, stream, batch_size=1):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    context.execute_async(batch_size=1, bindings=bindings,
                          stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


# Allocates all buffers required for an engine, i.e. host/device inputs/outputs
def allocate_buffers(engine):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        size = trt.volume(engine.get_binding_shape(binding))
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


# Simple helper data class that's a little nicer to use than a 2-tuple.
class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()
