# Copyright 2021-present Lenovo
# Confidential and Proprietary

import copy
import os
import sys

import numpy as np
import pycuda.autoinit  # noqa: F401
# This import causes pycuda to automatically manage CUDA context
# creation and cleanup.
import pycuda.driver as cuda  # noqa: F401
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorrt as trt
from trt_common import (allocate_buffers, convert_to_list, do_inference,
                        load_image, load_labels, output_to_csv, output_to_file)

# You can set the logger severity higher to suppress messages
# (or lower to display more messages).

parentpath = '/opt/letrain/'
path_list = [
    parentpath,
    parentpath + 'lib',
    parentpath + 'engine',
    parentpath + 'inference',
    parentpath + 'applications',
    os.path.join(parentpath, "applications/frcnn")
]
sys.path += path_list

from nets import nets_factory

tf.app.flags.DEFINE_string('plan_file', '', '')
tf.app.flags.DEFINE_string('label_file', '', '')
tf.app.flags.DEFINE_string('model_name', '', '')
tf.app.flags.DEFINE_string('input_dir', '', '')
tf.app.flags.DEFINE_string('output_dir', '', '')

FLAGS = tf.app.flags.FLAGS
TRT_LOGGER = trt.Logger(trt.Logger.WARNING)


class TensorRTPredictor(object):
    def __init__(
        self,
        model_name,
        plan_file,
        image_path,
        label_file='',
        output_path=''
    ):
        self.plan_file = plan_file
        self.label_file = label_file

        self.image_path = image_path
        self.output_path = output_path
        self.out_file = 'predict_result'

        self.model_name = model_name

    def _prepare(self):
        self.labels_dict = load_labels(self.label_file)
        num_classes = len(self.labels_dict)

        self.network_fn = nets_factory.get_network_fn(self.model_name,
                                                      num_classes,
                                                      is_training=False)
        self.image_size = self.network_fn.default_image_size

    def deserialize_cuda_engine(self):
        trt_runtime = trt.Runtime(TRT_LOGGER)
        with open(self.plan_file, "rb") as f:
            engine = trt_runtime.deserialize_cuda_engine(f.read())
        return engine

    def inference_by_engine(self, engine):  # noqa: C901
        batch_size = engine.max_batch_size
        img_arr_len = self.image_size * self.image_size * 3

        image_dict_list = []
        for filepath, dirnames, filenames in os.walk(self.image_path):
            for filename in filenames:
                image_dict_list.append({
                    "file_path": os.path.join(filepath, filename)
                })
        image_count = len(image_dict_list)

        with engine.create_execution_context() as context:
            inputs, outputs, bindings, stream = allocate_buffers(engine)

            for start_idx in range(0, image_count, batch_size):
                if start_idx + batch_size > image_count:
                    real_batch_size = image_count - start_idx
                else:
                    real_batch_size = batch_size

                source = load_image(
                    image_dict_list[start_idx]['file_path'], self.image_size
                )
                if real_batch_size > 1:
                    for i in range(start_idx + 1, start_idx + real_batch_size):
                        source = np.concatenate((
                            source,
                            load_image(
                                image_dict_list[i]['file_path'], self.image_size
                            )
                        ))
                    if len(source) < img_arr_len * batch_size:
                        source = np.concatenate((
                            source, np.zeros(
                                img_arr_len * (batch_size - real_batch_size)
                            )
                        ))
                np.copyto(inputs[0].host, source)
                [output] = do_inference(
                    context,
                    bindings=bindings,
                    inputs=inputs,
                    outputs=outputs,
                    stream=stream
                )
                batch_output_list = np.array_split(output, batch_size)
                for idx, item in enumerate(batch_output_list):
                    try:
                        image_dict_list[start_idx + idx]['post_value'] = \
                            copy.deepcopy(item)
                    except IndexError:
                        break
        return image_dict_list

    def inference(self):
        self._prepare()

        engine = self.deserialize_cuda_engine()
        image_dict_list = self.inference_by_engine(engine)

        predict_res = {}
        for image_file in image_dict_list:
            predict_idx = np.argmax(image_file['post_value'])
            predict_file = image_file['file_path'].replace(
                self.image_path.rstrip('/'),
                '.'
            )
            predict_res[predict_file] = self.labels_dict[str(predict_idx)]

        output_to_csv(
            os.path.join(self.output_path, self.out_file + '.csv'),
            ['image_path', 'predict_lable'],
            convert_to_list(predict_res)
        )
        output_to_file(
            predict_res,
            os.path.join(self.output_path, self.out_file + '.txt')
        )
        print(f"[INFO] Finished. You can check inference result file: \n"
              f"{os.path.join(self.output_path, self.out_file + '.csv')} \n"
              f"{os.path.join(self.output_path, self.out_file + '.txt')}")


if __name__ == "__main__":
    print(f"[INFO] Params: \n"
          f"model_name: {FLAGS.model_name} \n"
          f"plan_file: {FLAGS.plan_file} \n"
          f"label_file: {FLAGS.label_file} \n"
          f"image_path: {FLAGS.input_dir} \n"
          f"output_path: {FLAGS.output_dir}")
    predictor = TensorRTPredictor(
        model_name=FLAGS.model_name,
        plan_file=FLAGS.plan_file,
        label_file=FLAGS.label_file,
        image_path=FLAGS.input_dir,
        output_path=FLAGS.output_dir
    )

    predictor.inference()
