# Copyright 2022-present Lenovo
# Confidential and Proprietary

import argparse
import csv
import json
import os
import re
import sys
import time

re_dict = {}
parser_config = {}
re_map = {
    'time': r'.*(?P<time>\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\.\d+):.*',
    'epoch': r'.*epoch\s*(?P<epoch>\d+\.\d+|\d+),.*',
    'lr': r'.*lr\s*(?P<learning_rate>\d+\.\d+e-\d+),.*',
    'autodl_lr': r'.*lr\s*(?P<learning_rate>\d+\.\d+),.*',
    'loss': r'.*\s+loss\s*(?P<loss>\d+\.\d+|nan).*',
    'examples_per_sec': r'.*\((?P<examples_per_sec>\d+\.\d+)\s*examples/sec.*',
    'sec_per_batch': r'.*;\s*(?P<sec_per_batch>\d+\.\d+)\s*sec/batch.*',
    'delay': r'.*delay\s*(?P<delay>\d+\.\d+).*',
    'validation_loss': r'.*validation_loss\s(?P<loss>\d+\.\d+|nan).*',
    'top1_acc': r'.*top_1_accuracy\s*(?P<top1_acc>\d+\.\d+%),.*',
    'top5_acc': r'.*top_5_accuracy\s*(?P<top5_acc>\d+\.\d+%).*',
    'image_gan_training_loss': r'.*loss\s+\[(?P<generator_loss>.*?),\s+'
                               r'(?P<discriminator_loss>.*?)\].*',
    'image_gan_validation_loss': r'.*validation_loss\s+\[(?P<generator_loss>'
                                 r'.*?)\s+(?P<discriminator_loss>.*?)\].*',
    'meanAP': r'.*meanAP_results\s*(?P<meanAP>\d+\.\d+).*',
    'instance_epoch': r'\s*(?P<epoch>\d+/\d+)\s+\[.*?\].*',
    'instance_loss': r'.*ETA.*batch.*size.*loss:.*?(?P<loss>\d+\.\d+).*',
    'instance_val_loss': r'.*val_loss:\s*(?P<loss>\d+\.\d+).*'
}


def load_config():
    with open(os.path.join(sys.path[0], 'parser_config.json')) as f:
        parser_config.update(json.loads(f.read()))


def gen_dashboard_file():
    dashboard_content = template['dashboard']
    dashboard_path = os.path.join(args.workspace, 'dashboard.json')
    with open(dashboard_path, 'w') as f:
        f.write(json.dumps(dashboard_content))


def get_log_path():
    while True:
        time.sleep(1)
        log_paths = [f for f in os.listdir(args.workspace)
                     if f.endswith('.out') and args.outfile_identify in f]
        if not log_paths:
            continue
        return log_paths[0]


def update_to_csv(line_content):
    try:
        for record_file in template['record_file'].keys():
            regex = re_dict[record_file]
            re_result = regex.match(line_content)
            if not re_result:
                continue
            result_dict = re_result.groupdict()
            recoder_path = os.path.join(args.workspace, record_file)
            exists = os.path.exists(recoder_path)
            with open(recoder_path, 'a', newline='') as f:
                csv_writer = csv.writer(f)
                if not exists:
                    csv_writer.writerow(result_dict.keys())
                csv_writer.writerow(list(result_dict.values()))
    except Exception as e:
        print(f'Parse log error: {e}\n'
              f'current line is: {line_content}')


def gen_re_dict():
    re_dict.update({
        record_file: re.compile(''.join([re_map[field] for field in fields]))
        for record_file, fields in template['record_file'].items()
    })


def parse_log():
    with open(log_path, 'r') as f:
        while True:
            line_content = f.readline()
            if not line_content:
                time.sleep(1)
                continue
            if 'job end time is' in line_content:
                break
            update_to_csv(line_content)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--workspace', required=True)
    parser.add_argument('--job_template_code', required=True)
    parser.add_argument('--outfile_identify', required=True)
    args = parser.parse_args()

    try:
        load_config()
        if args.job_template_code not in parser_config:
            print(f'Parse log info: template {args.job_template_code}'
                  f' does not need to parse log')
            exit(0)

        log_path = get_log_path()
        print('Parse log info: successfully located to output file:', log_path)
        template = parser_config[args.job_template_code]
        gen_dashboard_file()
        print('Parse log info: dashboard.json was generated successfully')
        gen_re_dict()
        parse_log()
    except Exception as e:
        print('Parse log error:', e)
        exit(1)
